/***********************************************************************
 *
 * Copyright (C) 2008 Graeme Gott <graeme@gottcode.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#include "data_model.h"

#include <QApplication>
#include <QPalette>

#include <cmath>

/*****************************************************************************/

namespace {
	float hours(int val) {
		return std::floor(((val / 3600.f) + 0.05f) * 10.f) * 0.1f;
	}
}

/*****************************************************************************/

DataModel::DataModel(QObject* parent)
: QAbstractTableModel(parent) {
}

/*****************************************************************************/

bool DataModel::add(const QDateTime& start, const QDateTime& stop, bool billed) {
	if (stop < start) {
		return false;
	}

	if (start.date() == stop.date()) {
		// Find session position
		int pos = 0;
		foreach (const Session& session, m_data) {
			if (start < QDateTime(session.date, session.start)) {
				break;
			}
			pos++;
		}

		// Prevent intersecting sessions
		if (pos > 0) {
			const Session& session = m_data.at(pos - 1);
			if (start < QDateTime(session.date, session.stop)) {
				return false;
			}
		}
		if (pos < m_data.count()) {
			const Session& session = m_data.at(pos);
			if (stop > QDateTime(session.date, session.start)) {
				return false;
			}
		}

		// Insert session
		beginInsertRows(QModelIndex(), pos, pos);
		int duration = stop.toTime_t() - start.toTime_t();
		int daily = duration;
		int weekly = duration;
		int monthly = duration;
		if (pos < m_data.count() && !m_data.at(pos).billed) {
			const Session& session = m_data.at(pos);
			QDate date = session.date;
			if (start.date() == date) {
				daily += session.daily;
			}
			if (start.date().weekNumber() == date.weekNumber()) {
				weekly += session.weekly;
			}
			if (start.date().month() == date.month()) {
				monthly += session.monthly;
			}
		}
		Session session = { start.date(), start.time(), stop.time(), daily, weekly, monthly, false };
		m_data.insert(pos, session);
		endInsertRows();

		// Update successive billed sessions
		for (int i = 0; i < m_billed.count(); ++i) {
			int& session = m_billed[i];
			if (session >= pos) {
				session++;
			}
		}

		// Set billed status
		if (billed) {
			setBilled(pos, true);
		}

		// Increase totals for successive sessions
		updateTotals();
	} else {
		return add(start, QDateTime(start.date(), QTime(23, 59, 59)), billed) && add(QDateTime(stop.date(), QTime(0, 0, 0)), stop, billed);
	}

	return true;
}

/*****************************************************************************/

bool DataModel::edit(int pos, const QDateTime& start, const QDateTime& stop) {
	if (stop < start) {
		return false;
	}

	Q_ASSERT(pos < m_data.count());
	const Session& session = m_data.at(pos);
	QDateTime old_start = QDateTime(session.date, session.start);
	QDateTime old_stop = QDateTime(session.date, session.stop);

	// Replace session
	remove(pos);
	if (add(start, stop) == false) {
		add(old_start, old_stop);
		return false;
	} else {
		return true;
	}
}

/*****************************************************************************/

bool DataModel::remove(int pos) {
	if (pos >= m_data.count()) {
		return false;
	}

	// Remove session
	beginRemoveRows(QModelIndex(), pos, pos);
	m_data.removeAt(pos);
	endRemoveRows();

	// Update successive billed sessions
	m_billed.removeAll(pos);
	for (int i = 0; i < m_billed.count(); ++i) {
		int& session = m_billed[i];
		if (session > pos) {
			session--;
		}
	}

	return true;
}

/*****************************************************************************/

void DataModel::setBilled(int pos, bool billed) {
	Q_ASSERT(pos < m_data.count());

	if (!billed) {
		m_billed.removeAll(pos);
	} else {
		Q_ASSERT(!m_billed.contains(pos));
		m_billed.append(pos);
		qSort(m_billed);
	}
	m_data[pos].billed = billed;

	updateTotals();
}

/*****************************************************************************/

void DataModel::toXml(QXmlStreamWriter& xml) const {
	foreach (const Session& session, m_data) {
		xml.writeEmptyElement("session");
		xml.writeAttribute("date", session.date.toString(Qt::ISODate));
		xml.writeAttribute("start", session.start.toString(Qt::ISODate));
		xml.writeAttribute("stop", session.stop.toString(Qt::ISODate));
		xml.writeAttribute("billed", QString::number(session.billed));
	}
}

/*****************************************************************************/

int DataModel::rowCount(const QModelIndex& parent) const {
	if (!parent.isValid()) {
		return m_data.count();
	} else {
		return 0;
	}
}

/*****************************************************************************/

int DataModel::columnCount(const QModelIndex& parent) const {
	if (!parent.isValid()) {
		return 7;
	} else {
		return 0;
	}
}

/*****************************************************************************/

QVariant DataModel::data(const QModelIndex& index, int role) const {
	Q_ASSERT(index.isValid());
	if (role == Qt::DisplayRole) {
		const Session& session = m_data.at(index.row());
		switch (index.column()) {
		case 0:
			return session.date.toString("M/d/yy");
		case 1:
			return session.start.toString("h:mm A");
		case 2:
			return session.stop.toString("h:mm A");
		case 3:
			return hours(session.daily);
		case 4:
			return hours(session.weekly);
		case 5:
			return hours(session.monthly);
		default:
			return QVariant();
		}
	} else if (role == Qt::TextAlignmentRole) {
		return Qt::AlignRight;
	} else if (role == Qt::ForegroundRole && isBilled(index.row())) {
		return QApplication::palette().color(QPalette::Disabled, QPalette::Text);
	} else if (role == Qt::CheckStateRole && index.column() == 6) {
		return m_data.at(index.row()).billed ? Qt::Checked : Qt::Unchecked;
	} else {
		return QVariant();
	}
}

/*****************************************************************************/

Qt::ItemFlags DataModel::flags(const QModelIndex& index) const {
	return (index.column() != 6) ? QAbstractTableModel::flags(index) : (QAbstractTableModel::flags(index) |  Qt::ItemIsUserCheckable);
}

/*****************************************************************************/

QVariant DataModel::headerData(int section, Qt::Orientation orientation, int role) const {
	if (orientation != Qt::Horizontal) {
		return QVariant();
	}

	if (role == Qt::DisplayRole) {
		switch (section) {
		case 0:
			return tr("Date");
		case 1:
			return tr("Start");
		case 2:
			return tr("Stop");
		case 3:
			return tr("Daily");
		case 4:
			return tr("Weekly");
		case 5:
			return tr("Monthly");
		case 6:
			return tr("Billed");
		default:
			return QVariant();
		}
	} else {
		return QVariant();
	}
}

/*****************************************************************************/

bool DataModel::setData(const QModelIndex& index, const QVariant& value, int role) {
	if (index.column() == 6 && role == Qt::CheckStateRole) {
		setBilled(index.row(), value.toInt() == Qt::Checked);
		emit dataChanged(index, index);
		return true;
	} else {
		return QAbstractTableModel::setData(index, value, role);
	}
}

/*****************************************************************************/

void DataModel::updateTotals() {
	const Session& first = m_data.at(0);
	bool billed = first.billed;
	int daily = first.daily;
	int weekly = first.weekly;
	int monthly = first.monthly;

	for (int i = 1; i < m_data.count(); ++i) {
		const QDate& date = m_data.at(i - 1).date;
		Session& session = m_data[i];
		int duration = QDateTime(session.date, session.stop).toTime_t() - QDateTime(session.date, session.start).toTime_t();
		if (billed) {
			daily = weekly = monthly = duration;
		} else {
			if (session.date == date) {
				daily += duration;
			} else {
				daily = duration;
			}
			if (session.date.weekNumber() == date.weekNumber()) {
				weekly += duration;
			} else {
				weekly = duration;
			}
			if (session.date.month() == date.month()) {
				monthly += duration;
			} else {
				monthly = duration;
			}
		}
		session.daily = daily;
		session.weekly = weekly;
		session.monthly = monthly;
		billed = session.billed;
	}

	emit dataChanged(index(0, 0), index(m_data.count() - 1, 6));
}

/*****************************************************************************/
